
#ifndef _LIB_BURST_H_
#define _LIB_BURST_H_


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _OPENMP
#include <omp.h>
#endif


#include "../libraryStable/libraryBasic.h"
#include "../libraryStable/libsift.h"
#include "../libraryStable/libsiftIPOL.h"
#include "../libraryStable/libAsiftIPOL.h"
#include "poisson_fft.h"

extern "C" {
#include "triangle.h"
}




#define BagValueAfterRegistration -10000.0

#define INTERPOLATION_ORDER 1
//#define HOMOGRAPHY_RESAMPLING_ORDER INTERPOLATION_ORDER
//#define POLYNOMIAL_RESAMPLING_ORDER INTERPOLATION_ORDER
#define HOMOGRAPHY_RESAMPLING_ORDER 7
#define POLYNOMIAL_RESAMPLING_ORDER 7



/* Summary
 
 
 *** Group bursts and apply fusion for each burst
 
 1) Detect bursts in consecutive images
 Options: Ransac Parameters, degree of subsampling 
 
 
 2) Fusion each bursts
 Options: siftType, Ransac Parameters, distorsionFlag, adaptiveAveragingFlag
 
 

 *** Perform fusion with highlight removal
 Options: siftType, Ransac Parameters, distorsionFlag

 
 *** Perform grouping by scale and fusion with highlight removal
 Options: scales, siftType, Ransac Parameters, distorsionFlag
 

 
 */


/*  Algorithms
 
 *** Group bursts and apply fusion for each burst
 
 1) Perform consecutive bursts
 2) Subsample image of given factor
 3) Decide indexos of bursts by usual criteria 
 
 */




namespace libIIPStable 
{

    //! Auxiliary
    double max_dist_vertex(laMatrix &H, int iWidth, int iHeight);
  
    
    
    //! Main: group bursts
    void build_groups_of_images(cflimage *idata, int inImages, int &inBursts, int *fiIndexos, int *fiNumber, int iVerbose, int sfactor);

    void build_groups_of_images_lm(char **idata, int inImages, int &inBursts, int *fiIndexos, int *fiNumber, int iVerbose, int sfactor);
    
 
    
    //! Auxiliary
    void fit_polynomial(float *x0, float *y0, float *x1, float *y1, float *coeff_x, float *coeff_y, int n, int degree, int width, int height);
    void float_vector_apply_polynomial(float x0s, float y0s, float *coeff_x, float *coeff_y, int width, int height, float *xp, float *yp);    
    void apply_polynomial_picture(float *input, int width1, int height1, float *coeff_x, float *coeff_y, int spline_order, float *out, float x0, float y0, int nwidth, int nheight, int zoom);

    //! creates an image mask of the Delaunay triangulation: fLarge inside the convex hull; bg_here outside
    void set_triangulation_mask(flimage &mask, struct triangulateio out);
    void triangulate_Delaunay(float *x1, float *y1, int naccorded, flimage &mask);   
    
    
    
    
    //! Main: registration and reinterpolation of images with reference once
    // initIndex:               index of initial image
    // inImages:                number of consecutive images to be registered
    // iDistCorrectionFlag:     flag for using correction of distorsion
    // iAsifFlag:               flag for using Asift instead of sift
    // iVerbose:                verbose flag
    // iCorrectlyRegistered:    number of images correctly registered
    // ipCorrectlyRegistered:   indexos of correctly registered images
    // iCompressionFlag:        flag for restricting the compression factor in order to apply registration
    // lScale, uScale:          min, max Scales if iCompressionFlag activated
 	void register_burst_with_initial_image(cflimage *idata, cflimage *odata, int initIndex, int inImages, int iDistCorrectionFlag, int iDelaunayFlag,int iSiftType, int iVerbose, int & iCorrectlyRegistered, int *ipCorrectlyRegistered, int iCompressionFlag, float lScale, float uScale);

    
    
    
    void register_burst_with_initial_image_zoom(cflimage *idata, cflimage *odata, int initIndex, int inImages, int iDistCorrectionFlag, int iDelaunayFlag,int iSiftType, int iVerbose, int & iCorrectlyRegistered, int *ipCorrectlyRegistered, int iCompressionFlag, float lScale, float uScale, float fZoom);
    

    
    //! Main: perform fusion by averaging inside each burst
    // idata:           initial images
    // cfBursts:        output bursts
    // inBursts:        number of detected burst
    // fiIndex:         indexos of initial images of bursts
    // inImages:        total number of images
    // iAdaptiveBurst:  flag for performing adaptive averaging discarding blurred images
    void  average_already_registered_burst(cflimage *idata, cflimage *cfBursts, int inBursts, int * fiIndex, int inImages, int iAdaptiveBurst, int iVerbose);
    
    void  average_already_registered_single_burst(cflimage *idata, cflimage &cfBursts, int inImages, int iAdaptiveBurst, int iVerbose);
    
    
    //! Deprecated: Build groups and apply fusion in the same code
    //! Slower than making it separately 
    // void build_bursts(cflimage *idata, cflimage *odata, int inImages, int &inBursts, int *fiIndexos, int *fiNumber, int iVerbose);
	
    

    //! Auxiliar: perform specification of second image to first image histogram
    //! taking into account only coincident parts (images supposed to be registered)
    void histogram_specification_painting(float *in1, float *in2, float *out, int width1, int height1, int width2, int height2);  
    
    

    //! Main: combine registered burst
    //! All images are treated equivalently
    void  combine_registered_bursts_min(cflimage *idata, int inImages,  cflimage &cfCombined);
    void  combine_registered_bursts_medianRGB(cflimage *idata, int inImages,  cflimage &cfCombined);
    void  combine_registered_bursts_medianGrad(cflimage *idata, int inImages, cflimage &cfCombined);
    void  combine_registered_bursts_medianGrad_lm(cflimage *idata, int inImages, cflimage &cfCombined);

    
    void  combine_registered_bursts_medianGrad_dif_res(cflimage *idata, int inImages, cflimage &cfCombined);
    void  combine_registered_bursts_medianGrad_dif_res_lm(cflimage *idata, int inImages, cflimage &cfCombined);
    
    
    
}


/*

namespace libIIPStable 
{
	
    

    
    
 	
    
    
 
	
    
    
    
    
    
    
    
    
    
    
 void  combine_registered_bursts_medianRGB_dif_res( cflimage *idata, int inImages, cflimage &cfCombined);
 
   
    
    
    
    
}
*/




#endif
