#include "libdemosaicking.h"


namespace libIIPStable 
{
	
    
    
    
    
#define GREENPOSITION 0
#define REDPOSITION 1
#define BLUEPOSITION 2
    
    
    
    void demosaicking_init_cfa(int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height)
    {
        
        
        
        fpClear(ored,   0.0f, width*height);
        fpClear(ogreen, 0.0f, width*height);
        fpClear(oblue,  0.0f, width*height);
        
        
        // Initializations
        int bluex = 1 - redx;
        int bluey = 1 - redy;
        
        
        
        for(int x=0;x<width;x++)
            for(int y=0;y<height;y++)
            {
                
                if (x%2 == redx && y%2 == redy)         ored[y*width+x] = ired[y*width+x];
                else  if (x%2 == bluex && y%2 == bluey) oblue[y*width+x] = iblue[y*width+x];
                else                                    ogreen[y*width+x] = igreen[y*width+x];
                
            }
        
    
        
    }
    
    
    
    
    
    /**
     * @file   libdemosaicking.cpp
     * @brief  Demosaicking functions: HAmilton-Adams algorithm, NLmeans based demosaicking, Chromatic components filtering
     *
     * 
     *
     * @author Antoni Buades <toni.buades@uib.es>
     */
    
    
    
    
    /**
     * \brief  Classical Adams-Hamilton demosaicking algorithm
     *
     *  The green channel is interpolated directionally depending on the green first and red and blue second directional derivatives.
     *  The red and blue differences with the green channel are interpolated bilinearly.
     *
     * @param[in]  ired, igreen, iblue  original cfa image
     * @param[out] ored, ogreen, oblue  demosaicked output 
     * @param[in]  (redx, redy)  coordinates of the red pixel: (0,0), (0,1), (1,0), (1,1)
     * @param[in]  threshold value to consider horizontal and vertical variations equivalent and average both estimates
     * @param[in]  width, height size of the image
     *
     */
    
    
    
    void demosaicking_adams(float threshold, int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height)
    {
        
        
        
        fpCopy(ired,ored,width*height);
        fpCopy(igreen,ogreen,width*height);
        fpCopy(iblue,oblue,width*height);
        
        
        // Initializations
        int bluex = 1 - redx;
        int bluey = 1 - redy;
        
        
        // CFA Mask of color per pixel
        unsigned char* mask = (unsigned char *) malloc(width*height*sizeof(unsigned char));
        
        for(int x=0;x<width;x++)
            for(int y=0;y<height;y++){
                
                if (x%2 == redx && y%2 == redy) mask[y*width+x] = REDPOSITION;
                else  if (x%2 == bluex && y%2 == bluey) mask[y*width+x] = BLUEPOSITION;
                else 	mask[y*width+x] = GREENPOSITION;
                
            }
        
        
        
        
        // Interpolate the green channel by bilinear on the boundaries  
        // make the average of four neighbouring green pixels: Nourth, South, East, West
        for(int x=0;x<width;x++)
            for(int y=0;y<height;y++)
                if ( (mask[y*width+ x] != GREENPOSITION) && (x < 3 || y < 3 || x>= width - 3 || y>= height - 3 )) { 
                    
                    
                    int gn, gs, ge, gw;
                    
                    if (y > 0)  gn = y-1;	else    gn = 1;
                    if (y < height-1)	gs = y+1;  else  gs = height - 2 ;
                    if (x < width-1)  ge = x+1;  else  ge = width-2;
                    if (x > 0) gw = x-1;	else  gw = 1;
                    
                    ogreen[y*width + x] = (ogreen[gn*width + x] +  ogreen[gs*width + x] + ogreen[y*width + gw] +  ogreen[y*width + ge])/ 4.0;
                    
                }
        
        
        
        // Interpolate the green by Adams algorithm inside the image    
        // First interpolate green directionally
        for(int x=3;x<width-3;x++)
            for(int y=3;y<height-3;y++)
                if (mask[y*width+ x] != GREENPOSITION ) {  
                    
                    
                    int l = y*width+x;
                    int lp1 = (y+1)*width +x;
                    int lp2 = (y+2)*width +x;
                    int lm1 = (y-1)*width +x;
                    int lm2 = (y-2)*width +x;
                    
                    // Compute vertical and horizontal gradients in the green channel
                    float adv = fabsf(ogreen[lp1] - ogreen[lm1]);
                    float adh = fabsf(ogreen[l-1] - ogreen[l+1]);
                    float dh0, dv0;
                    
                    // If current pixel is blue, we compute the horizontal and vertical blue second derivatives
                    // else is red, we compute the horizontal and vertical red second derivatives
                    if (mask[l] == BLUEPOSITION){  
                        
                        dh0 = 2.0 * oblue[l] - oblue[l+2] - oblue[l-2];	
                        dv0 = 2.0 * oblue[l] - oblue[lp2] - oblue[lm2];
                        
                    } else { 			
                        
                        
                        dh0 = 2.0 * ored[l] - ored[l+2] - ored[l-2];	
                        dv0 = 2.0 * ored[l] - ored[lp2] - ored[lm2];
                        
                    }	
                    
                    // Add vertical and horizontal differences
                    adh = adh + fabsf(dh0);
                    adv = adv + fabsf(dv0);
                    
                    // If vertical and horizontal differences are similar, compute an isotropic average
                    if (fabsf(adv - adh) < threshold)
                        
                        ogreen[l] = (ogreen[lm1] +  ogreen[lp1] +  ogreen[l-1] + ogreen[l+1]) /4.0 + (dh0 + dv0) / 8.0;
                    
                    // Else If horizontal differences are smaller, compute horizontal average
                    else if (adh < adv )
                        
                        ogreen[l] = (ogreen[l-1] + ogreen[l+1])/2.0 + (dh0)/4.0;
                    
                    // Else If vertical differences are smaller, compute vertical average			
                    else if ( adv < adh ) 			
                        
                        ogreen[l] = (ogreen[lp1] + ogreen[lm1])/2.0 + (dv0)/4.0;
                    
                }
        
        
        // compute the bilinear on the differences of the red and blue with the already interpolated green
        demosaicking_bilinear_red_blue(redx,redy,ored,ogreen,oblue,width,height);
        
    }
    
    
    
    
    
    
    /**
     * \brief  Classical Adams-Hamilton demosaicking algorithm
     *
     *  The green channel is interpolated directionally depending on the green first and red and blue second directional derivatives.
     *  The red and blue differences with the green channel are interpolated bilinearly.
     *
     * @param[in]  ired, igreen, iblue  original cfa image
     * @param[out] ored, ogreen, oblue  demosaicked output 
     * @param[in]  (redx, redy)  coordinates of the red pixel: (0,0), (0,1), (1,0), (1,1)
     * @param[in]  threshold value to consider horizontal and vertical variations equivalent and average both estimates
     * @param[in]  width, height size of the image
     *
     */
    
    
    
    void demosaicking_adams_directional(int direction, int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height)
    {
        
        
        if (direction != 0 && direction != 1)
        {
            printf("... please select a vertical or horizontal direction for adams\n");
            exit(-1);
        }
        
        
        fpCopy(ired,ored,width*height);
        fpCopy(igreen,ogreen,width*height);
        fpCopy(iblue,oblue,width*height);
        
        
        // Initializations
        int bluex = 1 - redx;
        int bluey = 1 - redy;
        
        
        // CFA Mask of color per pixel
        unsigned char* mask = (unsigned char *) malloc(width*height*sizeof(unsigned char));
        
        for(int x=0;x<width;x++)
            for(int y=0;y<height;y++){
                
                if (x%2 == redx && y%2 == redy) mask[y*width+x] = REDPOSITION;
                else  if (x%2 == bluex && y%2 == bluey) mask[y*width+x] = BLUEPOSITION;
                else 	mask[y*width+x] = GREENPOSITION;
                
            }
        
        
        
        
        // Interpolate the green channel by bilinear on the boundaries  
        // make the average of four neighbouring green pixels: Nourth, South, East, West
        for(int x=0;x<width;x++)
            for(int y=0;y<height;y++)
                if ( (mask[y*width+ x] != GREENPOSITION) && (x < 3 || y < 3 || x>= width - 3 || y>= height - 3 )) { 
                    
                    
                    int gn, gs, ge, gw;
                    
                    if (y > 0)  gn = y-1;	else    gn = 1;
                    if (y < height-1)	gs = y+1;  else  gs = height - 2 ;
                    if (x < width-1)  ge = x+1;  else  ge = width-2;
                    if (x > 0) gw = x-1;	else  gw = 1;
                    
                    ogreen[y*width + x] = (ogreen[gn*width + x] +  ogreen[gs*width + x] + ogreen[y*width + gw] +  ogreen[y*width + ge])/ 4.0;
                    
                }
        
        
        
        // Interpolate the green by Adams algorithm inside the image    
        // First interpolate green directionally
        for(int x=3;x<width-3;x++)
            for(int y=3;y<height-3;y++)
                if (mask[y*width+ x] != GREENPOSITION ) {  
                    
                    
                    int l = y*width+x;
                    int lp1 = (y+1)*width +x;
                    int lp2 = (y+2)*width +x;
                    int lm1 = (y-1)*width +x;
                    int lm2 = (y-2)*width +x;
                    
                    // Compute vertical and horizontal gradients in the green channel
                    float adv = fabsf(ogreen[lp1] - ogreen[lm1]);
                    float adh = fabsf(ogreen[l-1] - ogreen[l+1]);
                    float dh0, dv0;
                    
                    // If current pixel is blue, we compute the horizontal and vertical blue second derivatives
                    // else is red, we compute the horizontal and vertical red second derivatives
                    if (mask[l] == BLUEPOSITION){  
                        
                        dh0 = 2.0 * oblue[l] - oblue[l+2] - oblue[l-2];	
                        dv0 = 2.0 * oblue[l] - oblue[lp2] - oblue[lm2];
                        
                    } else { 			
                        
                        
                        dh0 = 2.0 * ored[l] - ored[l+2] - ored[l-2];	
                        dv0 = 2.0 * ored[l] - ored[lp2] - ored[lm2];
                        
                    }	
                    
                    // Add vertical and horizontal differences
                    adh = adh + fabsf(dh0);
                    adv = adv + fabsf(dv0);
                    
                    
                    // If horizontal
                    if (direction == HORIZONTAL)
                        
                        ogreen[l] = (ogreen[l-1] + ogreen[l+1])/2.0 + (dh0)/4.0;
                    
                    // Else If vertical			
                    else if ( direction == VERTICAL) 			
                        
                        ogreen[l] = (ogreen[lp1] + ogreen[lm1])/2.0 + (dv0)/4.0;
                    
                    
                }
        
        
        // compute the bilinear on the differences of the red and blue with the already interpolated green
        demosaicking_bilinear_red_blue(redx,redy,ored,ogreen,oblue,width,height);
        
    }
    
    
    
    
    
    
    
    void md_cfim_variation1d(float *u,float *v,int direction, int nsize, int width, int height)
    {
        
        int x,y,i,s,l0;
        float sum,value0,value,suport;
        
        suport = (float) 2*nsize+1;
        
        if (direction == HORIZONTAL){ 		/* total variation along x axis */
            
            for(y=0;y<height;y++)
                for(x=0;x<width;x++){
                    
                    sum=0.0;
                    l0=y*width;
                    value0=u[l0+x];
                    for(i=-nsize;i<=nsize;i++){
                        s=x+i;
                        if (s>0 && s<width) value = (u[l0+s] - value0);
                        else value = 0.0;
                        sum+=value*value;
                    }
                    
                    v[l0+x] = sqrtf(sum /suport);
                }
            
        } else {
            
            for(y=0;y<height;y++)
                for(x=0;x<width;x++){
                    
                    sum=0.0;
                    value0=u[y*width+x];
                    for(i=-nsize;i<=nsize;i++){
                        s=y+i;
                        if (s>0 && s<height) value = (u[s*width+x] - value0);
                        else value = 0.0;
                        
                        sum+=value*value;
                    }
                    
                    v[y*width+x] = sqrtf(sum /suport);
                }
            
        }
        
    }
    
    
    
    
    
    
    void demosaicking_adams_chromatic_regularity(int side, int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height)
    {
        
        
        
        // horizontal interpolation
        float *arh = new float[width*height];
        float *agh = new float[width*height];
        float *abh = new float[width*height];
        
        
        demosaicking_adams_directional(HORIZONTAL, redx,redy, ired, igreen,iblue, arh, agh, abh,  width, height);
        
        
        
        // conversion to yuv of horizontal
        float *y = new float[width*height];
        float *u = new float[width*height];
        float *v = new float[width*height];
        
        fiRgb2Yuv(arh, agh, abh, y, u, v, width, height);
        
        
        
        // variation of horizontal	
        float *horuTv = new float[width*height];
        float *horvTv = new float[width*height];
        
        md_cfim_variation1d(u,horuTv,HORIZONTAL,side,width,height);
        md_cfim_variation1d(v,horvTv,HORIZONTAL,side,width,height);
        for (int i=0; i < width*height; i++) horuTv[i] += horvTv[i]; 
        
        
        
        
        
        
        
        // vertical interpolation
        float *arv = new float[width*height];
        float *agv = new float[width*height];
        float *abv = new float[width*height];
        
        demosaicking_adams_directional(VERTICAL, redx,redy, ired, igreen,iblue, arv, agv, abv,  width, height);
        
        
        
        
        // conversion to yuv of vertical
        fiRgb2Yuv(arv, agv, abv, y, u, v, width, height);
        
        
        
        // variation of vertical
        float *veruTv = new float[width*height];
        float *vervTv = new float[width*height];
        
        md_cfim_variation1d(u,veruTv,VERTICAL,side,width,height);
        md_cfim_variation1d(v,vervTv,VERTICAL,side,width,height);
        for (int i=0; i < width*height; i++) veruTv[i] +=vervTv[i]; 
        
        
        
        
        // combination of both
        
        for(int i=0;i<width*height;i++)
        {
            
            float pesTotal = veruTv[i] + horuTv[i];
            
            float pes;
            if (pesTotal>0.0)
                pes = veruTv[i]/ pesTotal;  
            else
                pes = 0.5;
            
            ored[i] = pes*arh[i] + (1.0-pes)* arv[i]; 
            ogreen[i] = pes*agh[i] + (1.0-pes)* agv[i]; 
            oblue[i] = pes*abh[i] + (1.0-pes) * abv[i]; 	  
            
        }
        
        
    }
    
    
    
    
    /**
     * \brief  Classical bilinear interpolation of red and blue differences with the green channel
     *
     *
     * @param[in]  ored, ogreen, oblue  original cfa image with green interpolated
     * @param[out] ored, ogreen, oblue  demosaicked output 
     * @param[in]  (redx, redy)  coordinates of the red pixel: (0,0), (0,1), (1,0), (1,1)
     * @param[in]  width, height size of the image
     *
     */
    
    void demosaicking_bilinear_red_blue(int redx,int redy,float *ored,float *ogreen,float *oblue,int width,int height)
    {
        
        
        
        //Initializations 
        int bluex = 1 - redx;
        int bluey = 1 - redy;
        
        
        // Mask of color per pixel
        unsigned char* mask = (unsigned char *) malloc(width*height*sizeof(unsigned char));
        
        for(int x=0;x<width;x++)
            for(int y=0;y<height;y++){
                
                if (x%2 == redx && y%2 == redy) mask[y*width+x] = REDPOSITION;
                else  if (x%2 == bluex && y%2 == bluey) mask[y*width+x] = BLUEPOSITION;
                else 	mask[y*width+x] = GREENPOSITION;
                
            }
        
        
        // Compute the differences  
        for(int i=0; i < width*height;i++){
            
            ored[i] -= ogreen[i];
            oblue[i] -= ogreen[i];
            
        }
        
        
        
        // Interpolate the blue differences making the average of possible values depending on the CFA structure 
        for(int x=0; x < width;x++)
            for(int y=0; y < height;y++)
                if (mask[y*width+x] != BLUEPOSITION){
                    
                    int gn, gs, ge, gw;
                    
                    // Compute north, south, west, east positions
                    // taking a mirror simmetry at the boundaries
                    if (y > 0)  gn = y-1;	else    gn = 1;
                    if (y < height-1)	gs = y+1;  else  gs = height - 2 ;
                    if (x < width-1)  ge = x+1;  else  ge = width-2;
                    if (x > 0) gw = x-1;	else  gw = 1;
                    
                    if (mask[y*width+x] == GREENPOSITION && y % 2 == bluey) 
                        oblue[y*width+x] = ( oblue[y*width+ge] + oblue[y*width+gw])/2.0;
                    else if (mask[y*width+x] == GREENPOSITION  && x % 2 == bluex) 
                        oblue[y*width+x] = ( oblue[gn*width+x] + oblue[gs*width+x])/2.0;
                    else {
                        oblue[y*width+x] =  (oblue[gn*width+ge] + oblue[gn*width + gw]  +  oblue[gs*width + ge] +  oblue[gs*width +gw])/4.0;		 
                    }
                    
                }
        
        
        
        
        // Interpolate the blue differences making the average of possible values depending on the CFA structure
        for(int x=0;x<width;x++)
            for(int y=0;y<height;y++)
                if (mask[y*width+x] != REDPOSITION){
                    
                    int gn, gs, ge, gw;
                    
                    // Compute north, south, west, east positions
                    // taking a mirror simmetry at the boundaries
                    if (y > 0)  gn = y-1;	else    gn = 1;
                    if (y < height-1)	gs = y+1;  else  gs = height - 2 ;
                    if (x < width-1)  ge = x+1;  else  ge = width-2;
                    if (x > 0) gw = x-1;	else  gw = 1;
                    
                    if (mask[y*width+x] == GREENPOSITION && y % 2 == redy) 
                        ored[y*width+x] = ( ored[y*width+ge] + ored[y*width+gw])/2.0;
                    else if (mask[y*width+x] == GREENPOSITION  && x % 2 == redx) 
                        ored[y*width+x] = ( ored[gn*width+x] + ored[gs*width+x])/2.0;
                    else {
                        ored[y*width+x] =  (ored[gn*width+ge] + ored[gn*width + gw]  +  ored[gs*width + ge] +  ored[gs*width +gw])/4.0;	 
                    }
                    
                }
        
        
        // Make back the differences
        for(int i=0;i<width*height;i++){
            
            ored[i] += ogreen[i];
            oblue[i] += ogreen[i];
        }
        
        
        
        
        
    }
    
    
    
    
    
    
    
    /**
     * \brief  NLmeans based demosaicking
     *
     * For each value to be filled, a weigthed average of original CFA values of the same channel is performed.
     * The weight depends on the difference of a 3x3 color patch 
     *
     * @param[in]  ired, igreen, iblue  initial demosaicked image
     * @param[out] ored, ogreen, oblue  demosaicked output 
     * @param[in]  (redx, redy)  coordinates of the red pixel: (0,0), (0,1), (1,0), (1,1)
     * @param[in]  bloc  research block of size (2+bloc+1) x (2*bloc+1)
     * @param[in]  h kernel bandwidth 
     * @param[in]  width, height size of the image
     *
     */
    
    
    void demosaicking_nlmeans(int bloc, float h,int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height)
    {
        
        
        // Initializations
        int bluex = 1 - redx;
        int bluey = 1 - redy;
        
        
        
        // CFA Mask of color per pixel
        unsigned char *cfamask = new unsigned char[width*height];
        
        
        for(int x=0;x<width;x++)
            for(int y=0;y<height;y++){
                
                if (x%2 == redx && y%2 == redy) cfamask[y*width+x] = REDPOSITION;
                else  if (x%2 == bluex && y%2 == bluey) cfamask[y*width+x] = BLUEPOSITION;
                else 	cfamask[y*width+x] = GREENPOSITION;
                
            }
        
        
        fpCopy(ired,ored,width*height);
        fpCopy(igreen,ogreen,width*height);
        fpCopy(iblue,oblue,width*height);
        
        
        
        // Tabulate the function Exp(-x) for x>0.
        int luttaille = (int) (LUTMAX*LUTPRECISION);
        float *lut = new float[luttaille];
        
        wxFillExpLut(lut, luttaille);
        
        
        
        
        // for each pixel
        for(int y=2; y <height-2; y++)
            for(int x=2; x<width-2; x++)
            {
                // index of current pixel
                int l=y*width+x;
                
                
                // Learning zone depending on the window size
                int imin=MAX(x-bloc,1);
                int jmin=MAX(y-bloc,1);
                
                int imax=MIN(x+bloc,width-2);
                int jmax=MIN(y+bloc,height-2);
                
                
                // auxiliary variables for computing average
                float red=0.0;
                float green=0.0;
                float blue=0.0;
                
                float rweight=0.0;
                float gweight=0.0;
                float bweight=0.0;
                
                
                // for each pixel in the neighborhood
                for(int j=jmin;j<=jmax;j++)
                    for(int i=imin;i<=imax;i++) {
                        
                        // index of neighborhood pixel
                        int l0=j*width+i;
                        
                        // We only interpolate channels differents of the current pixel channel
                        if (cfamask[l]!=cfamask[l0]) {
                            
                            
                            // Distances computed on color
                            float some = 0.0;
                            
                            some = fiL2FloatDist(ired,ired,x,y,i,j,1,1,width,width);
                            some += fiL2FloatDist(igreen,igreen,x,y,i,j,1,1,width,width);						
                            some += fiL2FloatDist(iblue,iblue,x,y,i,j,1,1,width,width);
                            
                            
                            
                            
                            
                            // Compute weight
                            some= some / (27.0 * h);
                            
                            float weight =  wxSLUT(some,lut);
                            
                            // Add pixel to corresponding channel average
                            
                            if (cfamask[l0] == GREENPOSITION)  {
                                
                                green += weight*igreen[l0];
                                gweight+= weight;
                                
                            } else if (cfamask[l0] == REDPOSITION) {
                                
                                red += weight*ired[l0];
                                rweight+= weight;
                                
                            } else {
                                
                                blue += weight*iblue[l0];
                                bweight+= weight;
                            }
                            
                        }
                        
                    }
                
                
                // Set value to current pixel
                if (cfamask[l] != GREENPOSITION && gweight > fTiny)  ogreen[l]  =   green / gweight; 
                else  ogreen[l] = igreen[l];
                
                if ( cfamask[l] != REDPOSITION && rweight > fTiny)  ored[l]  =  red / rweight ;
                else    ored[l] = ired[l];
                
                if  (cfamask[l] != BLUEPOSITION && bweight > fTiny)   oblue[l] =  blue / bweight;
                else  oblue[l] = iblue[l];
                
                
            } 
        
        delete[] cfamask;
        
    }
    
    
    
    
    
    
    
    void demosaicking_nlmeans_single_frame(int iIndex,	// iIndex of denoised frame
                                           int redx, int redy,	// red pixel position
                                           int iDtemp,	// Half size of temporal window
                                           int iDWin,		// Half size of comparison window  
                                           int iDBloc,		// Half size of research window
                                           float fFiltPar,	// Filtering parameter
                                           float ***fpI,		// Input
                                           float **fpO,		// Output
                                           int iFrames, int iChannels, int iWidth,int iHeight)
    {
        
        
        // length of each channel
        //	int iwxh = iWidth * iHeight;
        
        
        //  length of comparison window
        // int ihwl = (2*iDWin+1);	
        int iwl = (2*iDWin+1) * (2*iDWin+1);
        int icwl = iChannels * iwl;
        
        
        
        // filtering parameter
        // float fSigma2 = fSigma * fSigma;	
        float fH = fFiltPar;
        float fH2 = fH * fH;
        
        // multiply by size of comparison window, since distances are not normalized
        fH2 *= (float) icwl;
        
        
        
        // tabulate exp(-x), faster than using directly function expf
        int iLutLength = (int) rintf((float) LUTMAX * (float) LUTPRECISION);
        float *fpLut = new float[iLutLength];
        wxFillExpLut(fpLut,iLutLength);
        
        
        // initializations
        int bluex = 1 - redx;
        int bluey = 1 - redy;
        
        
        
        // CFA Mask of color per pixel
        unsigned char *cfamask = new unsigned char[iWidth*iHeight];
        
        
        for(int x=0;x<iWidth;x++)
            for(int y=0;y<iHeight;y++){
                
                if (x%2 == redx && y%2 == redy) cfamask[y*iWidth+x] = REDPOSITION;
                else  if (x%2 == bluex && y%2 == bluey) cfamask[y*iWidth+x] = BLUEPOSITION;
                else 	cfamask[y*iWidth+x] = GREENPOSITION;
                
            }
        
        
        
        
        
        
        // PROCESS STARTS
        // for each frame
        
#pragma omp parallel shared(fpI, fpO)
        for(int y=0; y < iHeight ; y++)
        {
            
            
#pragma omp for schedule(dynamic) nowait 
            
            for(int x=0 ; x < iWidth;  x++)
            {
                
                
                int l = y * iWidth + x;
                
                // reduce the size of the comparison window if we are near the boundary  
                int iDWin0 = MIN(iDWin,MIN(iWidth-1-x,MIN(iHeight-1-y,MIN(x,y))));
                
                
                // research zone depending on the boundary and the size of the window
                int tmin=MAX(iIndex-iDtemp,0);
                int tmax=MIN(iIndex+iDtemp,iFrames-1);
                
                int imin=MAX(x-iDBloc,iDWin0);
                int jmin=MAX(y-iDBloc,iDWin0);
                
                int imax=MIN(x+iDBloc,iWidth-1-iDWin0);
                int jmax=MIN(y+iDBloc,iHeight-1-iDWin0);
                
                
                
                
                // sum of weights
                float rweight=0.0;
                float gweight=0.0;
                float bweight=0.0;
                
                float red=0.0;
                float green=0.0;
                float blue=0.0;
                
                
                
                for(int t=tmin; t <= tmax; t++)
                    for(int j=jmin; j <= jmax; j++)
                        for(int i=imin ; i <= imax; i++) 
                            if (cfamask[l]!=cfamask[ j * iWidth + i]) 
                            {
                                
                                int l0 = j*iWidth+i;
                                
                                
                                float fDif = fiL2FloatDist(fpI[iIndex],fpI[t],x,y,i,j,iDWin0,iDWin0,iChannels,iWidth,iWidth);
                                fDif = fDif / fH2;	
                                
                                float fWeight = wxSLUT(fDif,fpLut);
                                
                                float **cfpI = fpI[t];
                                
                                if (cfamask[l0] == GREENPOSITION)  
                                {
                                    
                                    green += fWeight*cfpI[1][l0];
                                    gweight+= fWeight;
                                    
                                } else if (cfamask[l0] == REDPOSITION) 
                                {
                                    
                                    red += fWeight*cfpI[0][l0];
                                    rweight+= fWeight;
                                    
                                } else {
                                    
                                    blue += fWeight*cfpI[2][l0];
                                    bweight+= fWeight;
                                }
                                
                                
                                
                            }
                
                
                
                // Set value to current pixel
                float **cfpI = fpI[iIndex];
                if (cfamask[l] != GREENPOSITION && gweight > fTiny)  fpO[1][l]  =   green / gweight; 
                else  fpO[1][l] = cfpI[1][l];
                
                if ( cfamask[l] != REDPOSITION && rweight > fTiny)  fpO[0][l]  =  red / rweight ;
                else    fpO[0][l] = cfpI[0][l];
                
                if  (cfamask[l] != BLUEPOSITION && bweight > fTiny)   fpO[2][l] =  blue / bweight;
                else  fpO[2][l] = cfpI[2][l];
                
                
                
                
                
            }
            
            
        }
        
        
    }
    
    
    
    
    /**
     * \brief  Iterate median filter on chromatic components of the image
     *
     *
     * @param[in]  ired, igreen, iblue  initial  image
     * @param[in]  iter  number of iteracions
     * @param[out] ored, ogreen, oblue  filtered output 
     * @param[in]  (redx, redy)  coordinates of the red pixel: (0,0), (0,1), (1,0), (1,1)
     * @param[in]  side  median in a (2*side+1) x (2*side+1) window
     * @param[in]  projflag if not zero, values of the original CFA are kept 
     * @param[in]  width, height size of the image
     *
     */
    
    
    void chromatic_median(int iter,int redx,int redy,int projflag,float side,float *ired,float *igreen, float *iblue,float *ored,float *ogreen,float *oblue,int width,int height)
    {
        
        
        int size=height*width;
        
        // Auxiliary variables for computing chromatic components
        float *y=new float[size];
        float *u=new float[size];
        float *v=new float[size];
        float *u0=new float[size];
        float *v0=new float[size];
        
        int bluex=1-redx;
        int bluey=1-redy;
        
        
        // For each iteration
        for(int i=1;i<=iter;i++){
            
            
            // Transform to YUV
            fiRgb2Yuv(ired,igreen,iblue,y,u,v,width,height); 
            
            
            // Perform a Median on YUV component
            fiPatchMedian(u,u0,side,width,height); 
            fiPatchMedian(v,v0,side,width,height); 
            
            
            // Transform back to RGB
            fiYuv2Rgb(ored,ogreen,oblue,y,u0,v0,width,height); 
            
            // If projection flag activated put back original CFA values
            if (projflag)
            {
                
                for(int x=0;x<width;x++)
                    for(int y=0;y<height;y++){
                        
                        int l=y*width+x;
                        
                        if (x%2==redx && y%2==redy) ored[l]=ired[l];
                        
                        else if (x%2==bluex && y%2==bluey) oblue[l]=iblue[l];
                        
                        else ogreen[l]=igreen[l];
                        
                    }
                
            }
            
            
            fpCopy(ored,ired,size);
            fpCopy(ogreen,igreen,size);
            fpCopy(oblue,iblue,size);
            
        }
        
        
        // delete auxiliary memory
        delete[] y;
        delete[] u;
        delete[] u0;
        delete[] v;
        delete[] v0;
        
    }
    
    
    
    
    
    
    
    /**
     * \brief Demosaicking chain
     *
     *
     *
     * Compute initialization by Adams-Hamilton algorithm (u0)
     *
     * for h in {16,4,1} do 
     * {
     *
     *		u <- NL_h(u0);			Apply NLmeans demosaicking
     *
     *		u <- CR(u);				Apply chromatic regularization
     *
     *		u0 <- u;
     *
     * }
     *
     * Output <- u;
     *
     *
     * @param[in]  ired, igreen, iblue  initial  image
     * @param[out] ored, ogreen, oblue  filtered output 
     * @param[in]  (redx, redy)  coordinates of the red pixel: (0,0), (0,1), (1,0), (1,1)
     * @param[in]  width, height size of the image
     *
     */
    
    
    
    void ssd_demosaicking_chain(int redx,int redy,float *ired,float *igreen,float *iblue,float *ored,float *ogreen,float *oblue,int width,int height)
    {
        
        
        ////////////////////////////////////////////// Process
        
        
        
        float h;
        int dbloc = 7;
        float side = 1.5;
        int iter = 1;
        int projflag = 1;  
        float threshold = 2.0;
        
        
        demosaicking_adams(threshold,redx,redy,ired, igreen, iblue, ored, ogreen, oblue, width,height);
        
        
        
        h = 16.0;
        demosaicking_nlmeans(dbloc,h,redx,redy,ored,ogreen,oblue,ired,igreen,iblue,width,height);
        chromatic_median(iter,redx,redy,projflag,side,ired,igreen,iblue,ored,ogreen,oblue,width,height);
        
        
        
        h = 4.0;
        demosaicking_nlmeans(dbloc,h,redx,redy,ored,ogreen,oblue,ired,igreen,iblue,width,height);
        chromatic_median(iter,redx,redy,projflag,side,ired,igreen,iblue,ored,ogreen,oblue,width,height);
        
        
        
        h = 1.0;
        demosaicking_nlmeans(dbloc,h,redx,redy,ored,ogreen,oblue,ired,igreen,iblue,width,height);
        chromatic_median(iter,redx,redy,projflag,side,ired,igreen,iblue,ored,ogreen,oblue,width,height);
        
        
        
    }
    
    
}